/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 30-nov-2004
 */
package jasco.runtime;

import java.lang.annotation.Annotation;

/**
 * 
 * The thisJoinPoint keyword availabe in JAsCo aspects is of type MethodJoinpoint, this means that all subsequent methods are availabe for this keyword.
 * thisJoinPoint refers to the method joinpoint that triggered the aspect.
 * 
 * 
 * @author Wim
 *
 * 
 */
public interface MethodJoinpoint {
	
	/**
	 *  Returns java.lang.reflect.class objects denoting types of exceptions thrown by this method
	 */
	public abstract Class[] getExceptionTypes();
	
	
	/**
	 *  Returns java.lang.reflect.class object denoting the return type of this method
	 */
	public abstract Class getReturnType();

	
	/**
	 *  Returns the unique name for this method joinpoint
	 * 
	 * @deprecated use getSignature instead
	 */
	public abstract String getFullName();

	/**
	 *  Returns the unique name for this method joinpoint
	 */
	public abstract String getSignature();
	
	/**
	 *  Returns true when this method joinpoint is an execution joinpoint
	 */
	public abstract boolean isExecution();

	/**
	 *  Returns true when this method joinpoint is a call joinpoint
	 */
	public abstract boolean isCall();

	/**
	 *  Returns the method name
	 */
	public abstract String getName();
	
	/**
	 *  Loads this class
	 */
	public abstract Class loadClass();
	
	/**
	 *  Returns all defined annotations
	 */
	public abstract Annotation[] getAnnotations();
	
	/**
	 *  Returns the actual argument types
	 */
	public abstract Class[] getActualArgumentTypes();

	/**
	 *  Returns the formal argument types
	 */
	public abstract Class[] getFormalArgumentTypes();

	/**
	 *  Returns the fully quantified name of class the method is declared
	 */
	public abstract String getClassName();

	/**
	 *  Returns the modifiers intger, see java.lang.reflect.Modifier
	 */
	public abstract int getModifiers();

	/**
	 *  Returns an array of all arguments
	 */
	public abstract Object[] getArgumentsArray();

	/**
	 *  Returns the object the method is invoked upon
	 */
	public abstract Object getCalledObject();

	/**
	 *  Invokes the replaced method or subsequent around advices with original arguments, only availabe in around advice. 
	 *  Is equivalent to invoking proceed(), proceed is however the preferred because it will execute considerably faster because of optimizations.
	 * 
	 *  @throws ProceedFinishedException proceed was previously invoked and all advices and original method already executed.
	 */
	public abstract Object invokeJAsCoMethod() throws Exception;

	/**
	 *  Invokes the original method with original arguments, only availabe in around advice.
	 *  WARNING: This method bypasses the around chain, subsequent arounds are not executed! Use invokeJAsCoMethod if you just want
	 *  to proceed.
	 */
	public abstract Object invokeOriginalJAsCoMethod() throws Exception;

	/**
	 *  Invokes the replaced method or subsequent around advices, only availabe in around advice
	 *  @param obj The object to invoke the method upon, is typically thisJoinPointObject keyword
	 *  @param args The arguments to invoke the proceed chain with.
	 * 
	 * @throws ProceedFinishedException proceed was previously invoked and all advices and original method already executed.
	 */
	public abstract Object invoke(Object obj, Object[] args) throws Exception;

	/**
	 *  Invokes the original method, only availabe in around advice.
	 *  WARNING: This method bypasses the around chain, subsequent arounds are not executed! Use invokeJAsCoMethod if you just want
	 *  to proceed.
	 *  @param obj The object to invoke the method upon, is typically thisJoinPointObject keyword
	 *  @param args The arguments to invoke the original method with.
	 */
	public abstract Object invokeOriginal(Object obj, Object[] args)
			throws Exception;

	/**
	 * Allows to invoke the original method all over again with original arguments. This means that the list of applicable
	 * aspects is rebuild. WARNING: could easily cause an infinite loop if this aspect is applicable again! 
	 * Does not work with method overloading yet.
	 */
	public abstract Object invokeAgain() throws Exception;

	/**
	 * Allows to invoke the original method all over again. This means that the list of applicable
	 * aspects is rebuild and all aspects (including the current one) are triggered again if applicable. 
	 * WARNING: could easily cause an infinite loop if this aspect is applicable again! 
	 * Does not work with method overloading yet.
	 * @param calledobject The object to invoke the method upon, is typically thisJoinPointObject keyword
	 * @param args The arguments to invoke the original method with. 
	 */
	public abstract Object invokeAgain(Object calledobject, Object[] args) throws Exception;

	/**
	 * Allows to invoke the original method again with original arguments. If this aspect is however triggered by
	 * a method defined in another aspect (aspects on aspects), the triggering method for that aspect is used unless that aspect is triggered by a method 
	 * defined in an aspect etc...
	 */
	public abstract Object invokeNoAspectAgain() throws Exception;
	
	/**
	 * Returns the classLoader responsible for loading this joinpoint. When using reflection on certain classes related to the joinpoint use this classloader
	 * to ensure classloader safety in complicated class loader hierarchies.
	 */
	public ClassLoader getClassLoader();


	/**
	 * returns full class name of return type
	 * 
	 * @return full class name of return type
	 */
	public abstract String getReturnTypeName();


	/**
	 * returns array of full class names of formal arguments
	 * 
	 * @return full class names of formal arguments
	 */
	public abstract String[] getFormalArgumentTypeNames();

}