package jasco.tools.jascoparser;

/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
 
import jasco.options.Options;
import jasco.tools.aspectparser.PPackage;
import jasco.util.generators.MethodGenerator;
import jasco.util.generators.VariableGenerator;
import jasco.util.javacompiler.CompileError;
import jasco.util.logging.Logger;

import java.io.File;
import java.util.Iterator;
import java.util.Vector;

public class JascoParser implements JascoParserConstants
{
	private String filename;
    private int previousline;
	private JascoParserTokenManager tokenmanager;
    private Token currenttoken;
    private boolean isCheckingClasses =true;
    
    public static final int REFINABLEMOD =  0x08000000;
    
   
	public JascoParser(JascoParserTokenManager tokenmanager, String aFileName)  {
	
		previousline = 0; 
        this.tokenmanager = tokenmanager;
        filename = aFileName;
        currenttoken = tokenmanager.getNextToken();
	}
	
	public void setIsCheckingClasses(boolean b) {
		
		isCheckingClasses=b;
	}
	
	public boolean isCheckingClasses() {
		
		return isCheckingClasses;
	}
	 
	protected String getFileName()  {
	
		return filename;
	}
	
	protected void checkFileName(String name) throws JascoParseException {
		
		File file = new File(getFileName());
		String toCheck=file.getName();
		
		
		int index = toCheck.indexOf(".");
		if(index!=-1)
			toCheck=toCheck.substring(0,index);

		if(!toCheck.equals(name))
			throw new JascoParseException(this.newError("Name ("+name+") different than file name ("+toCheck+")!"));
	}
	
	protected int getPreviousLine()  {
	
		return previousline;
	}
	
	protected JascoParserTokenManager getTokenManager()  {
	
		return tokenmanager;
	}
	
	protected int getLine()
    {
        return tokenmanager.getLine();
    }

    protected void getNextToken() throws JascoParseException
    {
    	try {
        previousline = getLine();
        currenttoken = tokenmanager.getNextToken();
    	} catch(Throwable t) {throw new JascoParseException(newError(t.getMessage()));}
    }

    protected int getCurrentId()
    {
        return currenttoken.kind;
    }

    protected String getCurrentImage()
    {
        return currenttoken.image;
    }
	
	protected void assertToken(int givenToken, int expectedToken, CompileError error) throws JascoParseException
    {
        if (givenToken != expectedToken) 
		{
            throw new JascoParseException(error);
		}
    }

    protected void compareNotToken(String givenToken, String expectedToken, CompileError error) throws JascoParseException
    {
        if (givenToken.equals(expectedToken)) 
		{
            throw new JascoParseException(error);
		}
    }

    protected void assertOrToken(int givenToken, int expectedToken1, int expectedToken2, CompileError error) throws JascoParseException
    {
        if (givenToken != expectedToken1 && givenToken != expectedToken2) 
		{
            throw new JascoParseException(error);
		}
    }

    protected void assertNotToken(int givenToken, int expectedToken, CompileError error) throws JascoParseException
    {
        if (givenToken == expectedToken) 
		{
            throw new JascoParseException(error);
		}
    }
	
	protected CompileError newError(String desc)  {
	
		return new CompileError(getFileName(), getPreviousLine(), desc);
	}
	
	protected String generatePointer(int numberOfWhiteSpaces)
    {
        String result = "";
        for(int i = 0; i < numberOfWhiteSpaces; i++) 
        {
            result = result + " ";
        }
        return result + "^";
    }
	
	protected String checkClass(String name, PImports imports) throws JascoParseException {
		
		if(!this.isCheckingClasses())
			return name;
		
		String orig=name;
		if(hasWildcards(name))
			return name;
		if(isEvent(name))
			return name;	
		CompileError classNotFound = new CompileError(getFileName(), getPreviousLine(), "Class not found error: "+name);
		if(isPrimitive(name))
			return name;
		
		
		name = getFullClassName(name,imports);
		if(name!=null)
			return name;
	
		int index = orig.lastIndexOf(".");
		if (index!=-1)  {
			String newname = orig.substring(0,index)+"_"+orig.substring(index+1);
			if(!isClass(newname,imports))
				throw new JascoParseException(classNotFound);
			else return newname;
		}
		else throw new JascoParseException(classNotFound);
	}
	
	
	 protected PImports parseImportDeclaration() throws JascoParseException
		 {
		 	
		 	PImports pimports = new PImports();
		        while (getCurrentId() == IMPORT)
		        {
		            getNextToken();
		            String importname = getCurrentImage();
		            getNextToken();
					
		            String errorstring;
					
		            while (getCurrentId() == DOT)
		            {
		                getNextToken(); 
		                if(getCurrentId()!=CONNECTOR) {//connector is also package name, change this !!!
		                	errorstring = " <identifier> expected" + "\n\t" + "import " + importname + "." + "\n\t" + generatePointer(8 + importname.length());
		                	CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
		                	assertOrToken(getCurrentId(), IDENTIFIER , WILDCARDIDENTIFIER , error);
		                }
		                importname = importname + "." + getCurrentImage();	
						getNextToken();
		            }

		            pimports.addImport(new PImport(importname, getLine()));
					
		            errorstring = "import " + importname;
					checkSEMICOLON(errorstring); 
					
					getNextToken();		
		        }
		        return pimports;
		    }

		protected void checkSEMICOLONorLBRACE(String aErrorString) throws JascoParseException 
		{
			if ((getCurrentId() != SEMICOLON) && (getCurrentId() != LBRACE)) 
			{
				String errorstring = " ';' or '{' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
				CompileError error = new CompileError(filename, previousline, errorstring);
	            throw new JascoParseException(error);
			} 
		}
		
		protected void checkASSIGN(String aErrorString) throws JascoParseException
		{
			String errorstring = " '=' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
	        CompileError error = new CompileError(filename, previousline, errorstring);
	        assertToken(getCurrentId(), ASSIGN , error);
		}
		
		protected void checkNEW(String aErrorString) throws JascoParseException
		{
			String errorstring = " 'new' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
	        CompileError error = new CompileError(filename, previousline, errorstring);
	        assertToken(getCurrentId(), NEW , error);
		}

		
		protected void checkTYPEorLBRACE(String aErrorString) throws JascoParseException 
		{
		
		
			if (getCurrentId() != ANNOTATION && getCurrentId() != WILDCARDIDENTIFIER && getCurrentId() != STAR && getCurrentId() != BOOLEAN && getCurrentId() != BYTE && getCurrentId() != CHAR && getCurrentId() != DOUBLE && getCurrentId() != FLOAT && getCurrentId() != INT && getCurrentId() != LONG  && getCurrentId() != SHORT && getCurrentId() != VOID && getCurrentId() != IDENTIFIER && getCurrentId() != LBRACE) 
			{
				String errorstring = " 'type' or '{' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
	        	CompileError error = new CompileError(filename, previousline, errorstring);
	            throw new JascoParseException(error);
			}
		}
			
		
	protected void checkTYPE(String aErrorString) throws JascoParseException 
		{	
			if (getCurrentId() != WILDCARDIDENTIFIER && getCurrentId() != STAR && getCurrentId() != BOOLEAN && getCurrentId() != BYTE && getCurrentId() != CHAR && getCurrentId() != DOUBLE && getCurrentId() != FLOAT && getCurrentId() != INT && getCurrentId() != LONG  && getCurrentId() != SHORT && getCurrentId() != VOID && getCurrentId() != IDENTIFIER && getCurrentId() != PEROBJECT) 
			{
				String errorstring = " 'type' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
	        	CompileError error = new CompileError(filename, previousline, errorstring);
	            throw new JascoParseException(error);
			}
		}
		
	protected void checkDOT(String aErrorString) throws JascoParseException
		{
			String errorstring = " '.' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
	        CompileError error = new CompileError(filename, previousline, errorstring);
	        assertToken(getCurrentId(), DOT , error);
		}
		
	protected void checkLPAREN(String aErrorString) throws JascoParseException
    {
        String errorstring = " '(' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.LPAREN , error);
    }

    protected void checkRPAREN(String aErrorString) throws JascoParseException
    {
        String errorstring = " ')' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.RPAREN , error);
    }

    protected void checkLBRACE(String aErrorString) throws JascoParseException
    {
        String errorstring = " '{' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.LBRACE , error);
    }
    
    protected void checkRBRACKET(String aErrorString) throws JascoParseException
    {
        String errorstring = " ']' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.RBRACKET , error);
    }
    
    protected void checkLBRACKET(String aErrorString) throws JascoParseException
    {
        String errorstring = " '[' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.LBRACKET , error);
    }

    protected void checkRBRACE(String aErrorString) throws JascoParseException
    {
        String errorstring = " '}' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.RBRACE , error);
    }

    protected void checkSEMICOLON(String aErrorString) throws JascoParseException
    {
        String errorstring = " ';' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.SEMICOLON , error);
    }
	
	protected void checkGT(String aErrorString) throws JascoParseException
    {
        String errorstring = " '>' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.GT , error);
    }
	
	protected void checkCOLON(String aErrorString) throws JascoParseException
    {
        String errorstring = " ':' expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.COLON , error);
    }

    protected void checkIDENTIFIER(String aErrorString) throws JascoParseException
    {
        String errorstring = " <identifier> expected" + "\n\t" + aErrorString + "\n\t" + generatePointer(1 + aErrorString.length());
        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
        assertToken(getCurrentId(), JascoParserConstants.IDENTIFIER , error);
    }
    
    
    /*
    protected String accumulateTypeBeforeDot(String parsedType) throws JascoParseException {
    	
    	boolean starPrevious = false;
    	boolean begin=true;
    	if(parsedType.equals("*"))
    		starPrevious=true;
    	if(!parsedType.equals(""))
    		begin=false;
    	
    	while(getCurrentId()!=JascoParserConstants.DOT) {
    		System.out.println("start"+getCurrentImage());
    		if(isPrimitive(getCurrentImage())) {
    			parsedType += getCurrentImage();
    			getNextToken();
    			return parsedType;
    		}
    		else if((begin&&getCurrentId()==IDENTIFIER)||(starPrevious&&getCurrentId()==IDENTIFIER)) {
    			
    			begin=false;
    			starPrevious=false;
    			parsedType += getCurrentImage();
    			System.out.println("id"+parsedType);
    		}//aspect on aspect
    		else if((begin&&getCurrentId()==BEFORE)||(starPrevious&&getCurrentId()==BEFORE)) {
    			
    			begin=false;
    			starPrevious=false;
    			parsedType += getCurrentImage();
    			System.out.println("bf"+parsedType);
    		}
    		else if((begin&&getCurrentId()==AFTER)||(starPrevious&&getCurrentId()==AFTER)) {
    			
    			begin=false;
    			starPrevious=false;
    			parsedType += getCurrentImage();
    			
    		}
    		else if((begin&&getCurrentId()==REPLACE)||(starPrevious&&getCurrentId()==REPLACE)) {
    			
    			begin=false;
    			starPrevious=false;
    			parsedType += getCurrentImage();
  
    		}
    		else if(!begin&&!starPrevious&&getCurrentId()==STAR) {
    			begin=false;
    			starPrevious=true;
    			parsedType += getCurrentImage();
    			System.out.println("star"+parsedType);
    		}
    		else return parsedType;
    		getNextToken();
    	}
    	
    	return parsedType;
    }
	
	//accumulates type, ends at token after type, starts at first token (parsedType="") or first dot.(parsedType!="")
	protected String accumulateType(String parsedType)  throws JascoParseException
	{	
		System.out.println("FIRST"+parsedType);
		parsedType = accumulateTypeBeforeDot(parsedType);
		System.out.println("FIRST2"+parsedType);
		
		while(getCurrentId()==JascoParserConstants.DOT) {
			
			parsedType+=getCurrentImage();
			getNextToken();
			parsedType = parsedType+accumulateTypeBeforeDot("");
			
			if(getCurrentId()==LBRACKET)
    			break;
    		else if(getCurrentId()==LT)
				break;
		}
			
		if (getCurrentId()==JascoParserConstants.LT)  { //generics
			parsedType += getCurrentImage();
			getNextToken();
				
			while (true)  {
				parsedType+=accumulateType();
					
				if (getCurrentId()==JascoParserConstants.COMMA)  {
					parsedType+=getCurrentImage();
					getNextToken();
				}
				else  {
					checkGT(parsedType);
					parsedType+=getCurrentImage();
					getNextToken();
					break;
				}
			}			
		}		
		
		while(getCurrentId()==LBRACKET) { //arrays
			parsedType+=getCurrentImage();
			getNextToken();
			checkRBRACKET(parsedType);
			parsedType+=getCurrentImage();
			getNextToken();
		}
		System.out.println("LAST"+parsedType);
		
		return parsedType;
	}*/
    
    protected String accumulateType(PImports imports)  throws JascoParseException  {
		
		return accumulateType("",imports,TYPE_CLASS);
	}
    
    protected String accumulateType(String type, PImports imports)  throws JascoParseException  {
		
		return accumulateType(type,imports,TYPE_CLASS);
	}
    
    protected String accumulateType(PImports imports,int typeToParse)  throws JascoParseException  {
		
		return accumulateType("",imports,typeToParse);
	}
    
    public static final int TYPE_CLASS = 0;
    public static final int TYPE_METHOD = 1;
    public static final int TYPE_ASPECT = 2;
    
    /**
     * 
     * @param parsedType
     * @param imports
     * @param typeToParse is either CLASS, METHOD or ASPECT
     * @return
     * @throws JascoParseException
     */
    protected String accumulateType(String parsedType, PImports imports, int typeToParse)  throws JascoParseException
	{	
    	
		if (parsedType.equals(""))   { //start at first token
			parsedType += getCurrentImage();
			getNextToken();
		}
		else { 
			if (getCurrentId()==JascoParserConstants.DOT)  {
				parsedType += getCurrentImage();
				getNextToken();
				parsedType += getCurrentImage();
				getNextToken();
			}
			
			else return parsedType;
		} 
		
		boolean checked = false;
		
		while (true)
		{	
			
			if (getCurrentId()==JascoParserConstants.LT)  { //generics
				
				parsedType=checkAnyType(parsedType,imports,typeToParse);
				checked=true;
				parsedType+=getCurrentImage();
				getNextToken();
				
				while (true)  {
					parsedType+=accumulateType(imports);
					
					if (getCurrentId()==JascoParserConstants.COMMA)  {
						parsedType+=getCurrentImage();
						getNextToken();
					}
					else  {
						checkGT(parsedType);
						parsedType+=getCurrentImage();
						getNextToken();
						break;
					}
				}			
			}
			
			if(getCurrentId()==JascoParserConstants.DOT) {
				parsedType+=getCurrentImage();

				getNextToken();
				
				parsedType+=getCurrentImage();

				getNextToken();
			}
			else	break;
		}
		
		if(!checked) {
			parsedType=checkAnyType(parsedType,imports,typeToParse);
		}
		
		while(getCurrentId()==LBRACKET) { //arrays
			parsedType+=getCurrentImage();
			getNextToken();
			checkRBRACKET(parsedType);
			parsedType+=getCurrentImage();
			getNextToken();
		}

		return parsedType;
	}
    
    protected String checkAnyType(String parsedType, PImports imports, int typeToParse) throws JascoParseException {
    	
    	switch(typeToParse) {
		
			case TYPE_METHOD:
				int i = parsedType.lastIndexOf(".");
				if(i!=-1) {
					String type = parsedType.substring(0,i);
					parsedType=checkClass(type,imports)+parsedType.substring(i);
					}
				break; 
			
			case TYPE_CLASS:
				//if(parsedType.equals("void"))
				//	return parsedType;
				parsedType=checkClass(parsedType,imports);
				break;
				
			case TYPE_ASPECT:
				String parsedTypeTemp=isAspect(parsedType,imports);
				if(parsedTypeTemp==null) {
					CompileError classNotFound = new CompileError(getFileName(), getPreviousLine(), "Class not found error: "+parsedType);
					throw new JascoParseException(classNotFound);
				}
				else parsedType=parsedTypeTemp;
				break;
				
			default:
				throw new IllegalArgumentException("illegal parsed type");
		}
    	
    	return parsedType;
    }
    
    public String accumulateMethod(PImports imports) throws JascoParseException {
    	
    	return accumulateType(imports,TYPE_METHOD);
    }
	

	
	private String[] setBoe(String test) {
		return null;
	}
	
	public static boolean isRefinable(int i) {
		
			return (i & REFINABLEMOD)!=0;
	}
	
	public int setPublic(int modifiers) {
		
		modifiers|=java.lang.reflect.Modifier.PUBLIC;
		
		return modifiers;
	}
	
	protected void parseDeclarationPreamble(PJavaClass clz) throws JascoParseException {
		
		StringBuffer annotations = new StringBuffer();
    	int modifiers = parseModifiers(annotations,clz.getImports());
		
    	clz.setModifiers(modifiers);
    	clz.setAnnotations(annotations.toString());
	}
	
	protected int parseModifiers(StringBuffer annotations, PImports imports) throws JascoParseException {
		
		int mod = 0;
		
		while (true)  {
			
			int id = getCurrentId(); 
			switch (id)  {
			
			case JascoParserConstants.PUBLIC:
				mod|= java.lang.reflect.Modifier.PUBLIC;
				break;
				
			case JascoParserConstants.PROTECTED:
				mod|= java.lang.reflect.Modifier.PROTECTED;
				break;
			
			case JascoParserConstants.STATIC:
				mod|= java.lang.reflect.Modifier.STATIC;
				break;
			
			case JascoParserConstants.STRICTFP:
				mod|= java.lang.reflect.Modifier.STRICT;
				break;
			
			case JascoParserConstants.VOLATILE:
				mod|= java.lang.reflect.Modifier.VOLATILE;
				break;
				
			case JascoParserConstants.TRANSIENT:
				mod|= java.lang.reflect.Modifier.TRANSIENT;
				break;
			
			case JascoParserConstants.NATIVE:
				mod|= java.lang.reflect.Modifier.NATIVE;
				break;
				
			case JascoParserConstants.SYNCHRONIZED:
				mod|= java.lang.reflect.Modifier.SYNCHRONIZED;
				break;
				
			case JascoParserConstants.PRIVATE:
				mod|= java.lang.reflect.Modifier.PRIVATE;
				break;
				
				
			
			case JascoParserConstants.ABSTRACT:
				mod|= java.lang.reflect.Modifier.ABSTRACT;
				break;
				
			case JascoParserConstants.FINAL:
				mod|= java.lang.reflect.Modifier.FINAL;
				break;
				
			case JascoParserConstants.REFINABLE:
				mod|= REFINABLEMOD;
				break;
				
			case JascoParserConstants.ANNOTATION:
				parseAnnotation(annotations,imports);
				break;
				
			default:
				return mod;
			}
			if(id!=JascoParserConstants.ANNOTATION) 
				getNextToken();
		}
	}
	
	protected void parseAnnotation(StringBuffer annotations, PImports imports) throws JascoParseException {
		
		assertToken(getCurrentId(),ANNOTATION,newError("annotation (@) token expected \n"+annotations.toString()));
		annotations.append(getCurrentImage());

		getNextToken();
		checkIDENTIFIER(annotations.toString());
		
		annotations.append(accumulateType(imports));
		
		if(getCurrentId()!=LPAREN) {
			annotations.append(" ");
			return;
		}
		
		annotations.append(getCurrentImage());
		getNextToken();
		
		while(true) {
			annotations.append(getCurrentImage());
			if(getCurrentId()==RPAREN)
				break;
			else if(getCurrentId()==LBRACE) {
				getNextToken();
				parseAnnotation(annotations,imports);
				while(getCurrentId()==COMMA) {
					annotations.append(getCurrentImage());
					getNextToken();
					parseAnnotation(annotations,imports);
				}
			}
			else getNextToken();
		}
		
		getNextToken();
	}
	
	protected String removeArrayStuff(String clz) {
		
		int i = clz.indexOf("[");
		if(i!=-1)
			return clz.substring(0,i);
		else return clz;
	}
	
	protected String addArrayStuff(String clz, String orig) {
		
		int i = orig.indexOf("[");
		if(i!=-1) {
			String n = orig.substring(i);
			return clz+n;
		}
		else return clz;
		
	}
	
	protected boolean isClass(String className, PImports imports)  {
		
		
		String name = getFullClassName(className,imports);
		if(name==null) return false;
		else {
			return Options.isClass(name);
		}
			
		/*
		className=removeArrayStuff(className);
		
		if(Options.isClass(className))
			return true;
		 
		Vector packages = new Vector();
		Iterator i  = imports.getImports();
		while (i.hasNext())  {
			PImport next = (PImport) i.next();
			String pack =  next.getImportName();
			if(pack.endsWith("*"))
				packages.add(pack.substring(0,pack.length()-1));
			else if (pack.endsWith(className)) 	 
				return true;
		}
		//packages.add(cutpointClass.getPackageName()+".");
		packages.add("java.lang.");
		
		i = packages.iterator();
		while (i.hasNext())  {
			String next = i.next().toString();
			try {
			//System.out.println("testing:"+next+currentName);
			Options.loadClass(next+className);
			return true;
			}
			catch(ClassNotFoundException ex) {
			}
		}
		
		return false;*/
	}
	
	
	private String[] primitives = new String[]  {"int","float","double","long","short","byte","char","boolean","void"};
	protected boolean isPrimitive(String aClass) 
	{
		for(int i=0;i<primitives.length;++i)
			if(aClass.equals(primitives[i]))
				return true;
		
		return false;	
	
	}
	
	protected String getFullClassName(String origClassName, PImports imports)  {
		
		if(!this.isCheckingClasses())
			return origClassName; 
		
		if(isGenerics(origClassName))
			return origClassName;
		
		String className=removeArrayStuff(origClassName);
		
		if(Options.isClass(className))
			return addArrayStuff(className,origClassName);	
		 
		Vector packages = new Vector();
		Iterator i  = imports.getImports();
		while (i.hasNext())  {
			PImport next = (PImport) i.next();
			String pack =  next.getImportName();
			if(pack.endsWith("*"))
				packages.add(pack.substring(0,pack.length()-1));
			else if (pack.endsWith("."+className)||pack.equals(className)) 	{
				if(Options.isClass(pack))
					return addArrayStuff(pack,origClassName);
			}
		}
		//packages.add(cutpointClass.getPackageName()+".");
		packages.add("java.lang.");
		
		String result=null;
		
		i = packages.iterator();
		while (i.hasNext())  {
			String next = i.next().toString();
			try {
			
			Options.loadClass(next+className);
			if(result!=null) 
				Logger.getInstance().showError("Warning: Ambigious class reference: "+className);
			else result=next+className;
			}
			catch(ClassNotFoundException ex) {
			}
		}
		
		return addArrayStuff(result,origClassName);
	}
	

	protected PJavaCode parseMethodBody() throws JascoParseException {
		
		
		
		PJavaCode code = new PJavaCode();
	    int cachedlinenumber = getLine();
	    String temp = "";
		int numberofleftbraces = 1;
		
	    while (numberofleftbraces!=0)
	    {
	        String errorstring = "unexpected end of file";
	        CompileError error = new CompileError(getFileName(), getPreviousLine(), errorstring);
	        assertNotToken(getCurrentId(), EOF, error);
	       
	        if(getCurrentId()==DOT||getCurrentId()==COMMA||getCurrentImage().startsWith(".")) { //fe numbers: 0.9 (0 .9  is invalid...)
	    
	        	if(temp.length()>0&&temp.charAt(temp.length()-1)==' ') {
	        		temp=temp.substring(0,temp.length()-1);
	        	}
	        }
			
	        if (getCurrentId() == LBRACE) 
			{
	           numberofleftbraces++;
			}
	        else if (getCurrentId() == RBRACE) 
			{	
	            numberofleftbraces--;
			}
			
			if(numberofleftbraces!=0) 
				temp = temp + getCurrentImage() + " ";
			
	        getNextToken();
			
	        if(cachedlinenumber != getLine())
	        {
	            code.addJavaCode(temp, cachedlinenumber);
	            cachedlinenumber = getLine();
	            temp = "";
	        }
	    }
	    return code;
	}

	protected void parseMethod(MethodGenerator method, PImports imports) throws JascoParseException {
		
		getNextToken();
		while (getCurrentId()!=RPAREN)  {
			String vtype = accumulateType(imports);
			
			checkIDENTIFIER(getCurrentImage());
			
			String vname = getCurrentImage();
			
			method.addVariable(vtype,vname);
			
			getNextToken();
			
			assertOrToken(getCurrentId(),COMMA,RPAREN,newError(", or ) expected"));
			if(getCurrentId()==COMMA)
				getNextToken();
		}
		 
		getNextToken();
		if (getCurrentId()==THROWS)  {
			getNextToken();
			
			while (true)  {
				String etype = accumulateType(imports);
				method.addException(etype);
				if(getCurrentId()!=COMMA)
					break;
				else getNextToken(); 
			} 
		}
		
		if (getCurrentId()==SEMICOLON&&(java.lang.reflect.Modifier.isAbstract(method.getModifiers())||isRefinable(method.getModifiers()))) { //abstract method
			 getNextToken();
		}
		else  {
			
			checkLBRACE(method.getName());
			getNextToken();
			PJavaCode code = parseMethodBody();
		
			method.setImplementation(code.toString());
		}
	}

	protected void parseField(VariableGenerator field) throws JascoParseException {
		
		assertOrToken(getCurrentId(),ASSIGN,SEMICOLON,newError(" = or ; expected"));
		
		if (getCurrentId()==SEMICOLON)  {
			getNextToken();
			return;
		}
		else  {
			getNextToken();
			StringBuffer code = new StringBuffer();
			while (getCurrentId()!=SEMICOLON)  {
				code.append(getCurrentImage()+" ");
				getNextToken();
			}
		
			field.setInitialiser(code.toString());
		}	
		
		getNextToken();	
	}

	protected void parseJavaCode(PJavaClass parent, PImports imports) throws JascoParseException {	
		StringBuffer annotations = new StringBuffer();
		int modifiers = parseModifiers(annotations, imports);
		
		parseJavaCode(parent,imports,annotations,modifiers);
	}
	
	protected void parseSuperDeclarations(PJavaClass target, PImports imports) throws JascoParseException {
		
		if(getCurrentId()==EXTENDS) {
			getNextToken();
			checkIDENTIFIER("super class name expected for"+target.getName());
			String type = accumulateType(imports);
			target.setSuperClass(type);
		}
		if(getCurrentId()==IMPLEMENTS) {
			getNextToken();
			checkIDENTIFIER("super interface name expected for"+target.getName());
			String type = accumulateType(imports);
			target.addInterface(type);
			while(getCurrentId()==COMMA) {
				getNextToken();
				checkIDENTIFIER("super interface name expected for"+target.getName());
				String itype = accumulateType(imports);
				target.addInterface(itype);
			}
		}
	}
	
	protected void parseClass(PJavaClass parent, PImports imports, StringBuffer annotations, int modifiers) throws JascoParseException {
		
		checkIDENTIFIER("class name expected");
		
		String classname = getCurrentImage();
		PJavaClass innerClass = new PJavaClass(classname);
		getNextToken();
		
		parseSuperDeclarations(innerClass,imports);
		
		checkLBRACE("opening brace expected for class");
		getNextToken();
		
		while(getCurrentId()!=RBRACE) {
			parseJavaCode(innerClass,imports);
		}
		
		parent.addInnerClass(innerClass);
		
		getNextToken();
		
	}
	
	
	
	protected void parseJavaCode(PJavaClass parent, PImports imports, StringBuffer annotations, int modifiers) throws JascoParseException {	
		
		if(getCurrentId()==CLASS) {
			
			getNextToken();
			parseClass(parent,imports,annotations,modifiers);
			return;
		}
			
		else if (getCurrentId()==LBRACE)   { //static init
			getNextToken();
			PJavaCode code = parseMethodBody();
			parent.addStaticInitializer(code);
			return; 
		}
		
		String type = accumulateType(imports);
		
		checkIDENTIFIER(getCurrentImage());
		
		String name = getCurrentImage(); 
		
		getNextToken();
		
		if(getCurrentId()==LPAREN)  { //method
			MethodGenerator method = new MethodGenerator(name,type);
			method.addAnnotation(annotations.toString());
			PMethod pm = new PMethod(method,getPreviousLine());
			method.setModifiers(modifiers);
			pm.setLine(getPreviousLine());
			parent.addMethod(pm);
			parseMethod(method,imports);
		}
		else  { 
			VariableGenerator field = new VariableGenerator(type,name);
			PField pf = new PField(field,getPreviousLine());
			field.addAnnotation(annotations.toString());
			field.setModifiers(modifiers);
			pf.setLine(getPreviousLine());
			parent.addField(pf);
			parseField(field);
		}
	}
	
	protected String isAspect(String className, PImports imports) {
		
		if(!this.isCheckingClasses())
			return className;
		
		Vector packages = new Vector();
		Iterator i =imports.getImports();
		while (i.hasNext()) {
			PImport next = (PImport) i.next();
			String pack = next.getImportName();
			if (pack.endsWith("*"))
				packages.add(pack.substring(0, pack.length() - 1));
			else if (pack.endsWith("."+className)||pack.equals(className))
				if (Options.isAspect(pack))
					return pack;
				else
					return null;
		}
		//packages.add(cutpointClass.getPackageName()+".");
		packages.add("java.lang.");
		packages.add("");

		i = packages.iterator();
		while (i.hasNext()) {
			String next = i.next().toString();
			String newTest = next + className;
			//System.out.println("testing:"+next+currentName);
			//Options.loadClass(next+className);
			boolean b = Options.isAspect(newTest);
			if (b)
				return newTest;
			
			int index = newTest.lastIndexOf(".");
			if(index!=-1) {
				newTest=newTest.substring(0,index)+"_"+newTest.substring(index+1);
				
				b = Options.isAspect(newTest);
				if (b) 
					return newTest;
				
			}
			
		}

		return null;
	}

	protected boolean hasWildcards(String name) {
	
		if((name.indexOf("*")!=-1)||(name.indexOf("?")!=-1))
			return true;
		else return false;
	}
	
	protected boolean isArray(String s) {
		
		return s.indexOf("[")!=-1;
	}
	
	protected boolean isGenerics(String s) {
		
		return s.indexOf("<")!=-1;
	}
	
	public static String getClassCode(String aClass) throws ClassNotFoundException {
		
		if(aClass.equals("boolean"))
			return "Z";
		else if(aClass.equals("int"))
			return "I";
		else if(aClass.equals("float"))
			return"F";
		else if(aClass.equals("double"))
			return "D";
		else if(aClass.equals("char"))
			return "C";
		else if(aClass.equals("byte"))
			return "B";
		else if(aClass.equals("short"))
			return "S";
		else if(aClass.equals("long"))
			return "J";
		else return "L"+aClass+";";
	}
	
	protected Class loadClassOrPrim(String aClass) throws ClassNotFoundException {
		
		if(aClass.equals("boolean"))
			return Boolean.TYPE;
		else if(aClass.equals("int"))
			return Integer.TYPE;
		else if(aClass.equals("float"))
			return Float.TYPE;
		else if(aClass.equals("double"))
			return Double.TYPE;
		else if(aClass.equals("char"))
			return Character.TYPE;
		else if(aClass.equals("byte"))
			return Byte.TYPE;
		else if(aClass.equals("short"))
			return Short.TYPE;
		else if(aClass.equals("long"))
			return Long.TYPE;
		else if(aClass.equals("void"))
			return Void.TYPE;	
		else return Options.loadClass(aClass);
	}

	protected Class loadClassOrPrimOrArray(String aClass) throws ClassNotFoundException {
		
		if(isArray(aClass)) {
			String classname = removeArrayStuff(aClass);
			classname = getClassCode(classname);
			int i = aClass.indexOf("[");
			if(i!=-1) {
				int n = (aClass.length()-i)/2;
				for(int j=0;j<n;++j)
					classname="["+classname;
			}
			return Options.loadClass(classname);
		}
		
		return loadClassOrPrim(aClass);
		
		
	}

	protected boolean isEvent(String name) {
	
		return name.equals("onevent");
	}

	protected void parsePackageDeclaration(PJavaClass target) throws JascoParseException {
	    if(getCurrentId() == PACKAGE)
	    {
	        getNextToken();
	        String packagename = getCurrentImage();
	        getNextToken();
			
	        String errorstring;
			
	        while (getCurrentId() == DOT) 
	        {
	            getNextToken();
	            errorstring = "package " + packagename + ".";
	            checkIDENTIFIER(errorstring);  
	            packagename = packagename + "." + getCurrentImage();
				getNextToken();
	        }
	
	        target.setPackage(new PPackage(packagename, getLine()));
			
	        errorstring = "package " + packagename;
	        checkSEMICOLON(errorstring); 
	        
			getNextToken();
	    }
	}
	
}
	