/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util.generators;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import java.util.Vector;

public class MethodGenerator extends JavaGenerator {
    private String       returnType;
    private final Vector variableNames = new Vector();
    private final Vector variableTypes = new Vector();
    private String       implementation = "";
    private final Vector exceptions = new Vector();
    
    public final String ARGS = "$$";

    public MethodGenerator(String name, String returnType) {
        super(name);
        this.returnType = returnType;
    }
    
    public MethodGenerator copy(String newname) {
    	
    	MethodGenerator g = new MethodGenerator(newname,returnType);
    	g.variableNames.addAll(this.variableNames);
    	g.variableTypes.addAll(this.variableTypes);
    	g.implementation=this.implementation;
    	g.exceptions.addAll(this.exceptions);
    	g.setModifiers(this.getModifiers());
    	return g;
    }
    
    public int getVariableCount() {
    	
    	return variableNames.size();
    }
    
    public String generateInvocation() {
    	
    	String res="";
    	res+=getName()+"(";
    	Iterator i = variableNames.iterator();
  
    	while(i.hasNext()) {
    		String name = i.next().toString();
    		res+=name;
    		if(i.hasNext())
    			res+=", ";
    	}
    	
    	res+=")";
    	
    	return res;
    }

    public String getReturnType() {
        return returnType;
    } 
	
	public void setReturnType(String type)  {
	
		this.returnType=type;
	}

    public void addVariable(String type, String name) {
        variableTypes.add(type);
        variableNames.add(name);
    } 
    
    public void insertVariable(String type, String name) {
        variableTypes.insertElementAt(type,0);
        variableNames.insertElementAt(name,0);
    } 

    public Iterator getVariableNames() {
        return variableNames.iterator();
    } 

    public Iterator getVariableTypes() {
        return variableTypes.iterator();
    } 

    public void setImplementation(String implementation) {
    	
        this.implementation = implementation;
    } 

    public String getImplementation() {
        return implementation;
    } 

    public boolean hasImplementation() {
        return !implementation.equals("");
    } 

    public void addException(String e) {
        exceptions.add(e);
    } 

    public Iterator getExceptions() {
        return exceptions.iterator();
    } 

    public boolean hasExceptions() {
        return exceptions.size() > 0;
    }
    
    public String toString() {
        String code = super.toString();
        code+=getAnnotationsString();
        if (Modifier.isPublic(getModifiers()))
            code += "public ";
        else if (Modifier.isPrivate(getModifiers()))
            code += "private ";
        else if (Modifier.isProtected(getModifiers()))
            code += "protected ";
        
        if (Modifier.isAbstract(getModifiers()))
         	code += "abstract ";

        if (Modifier.isStatic(getModifiers()))
            code += "static ";

        if (Modifier.isFinal(getModifiers()))
            code += "final ";
        
        if (Modifier.isSynchronized(getModifiers()))
         	code += "synchronized ";

        code += getReturnType() + " ";
        code += getName() + "(";

        Iterator v = getVariableNames();
        Iterator t = getVariableTypes();
        if (t.hasNext()) {
            String name = (String) v.next();
            String type = (String) t.next();
            code += type + " " + name;

            while (t.hasNext()) {
                name = (String) v.next();
                type = (String) t.next();
                code += ", " + type + " " + name;
            } 
        } 

        code += ") ";

        Iterator ex = getExceptions();
        if (ex.hasNext()) {
            code += "throws ";

            String exception = (String) ex.next();
            code += exception;

            while (ex.hasNext()) {
                exception = (String) ex.next();
                code += ", " + exception;
            } 
        } 

        if (hasImplementation()) {
        	 if(hasLineNo())
            	code+=" //"+getLineNo()+NEWLINE;
        	code += " {" + NEWLINE + NEWLINE + addTabsBeforeLines(generateImplementation()) + NEWLINE + "}";
        }
        else
            code += ";";
        
        if(hasLineNo())
        	code+=" //"+getLineNo();

        return code;
    } 
    
    protected String generateImplementation() {
    	
    	String s = getImplementation();
    	
    	int i = s.indexOf(ARGS);
    	String args = "";
    	if(i>=0) {
    		Iterator it = getVariableNames();
    		while(it.hasNext()) {
    			args+=it.next().toString();
    			if(it.hasNext())
    				args+=", ";
    		}
    	}
    	while(i>=0) {
    		s = s.substring(0,i)+args+s.substring(i+ARGS.length());
    		i = s.indexOf(ARGS,i+1);
    	}
    	
    	return s;
    }

	/**
	 * @param next
	 * @return
	 */
	public boolean hasSameSignatureAs(Method next) {
	
		if(!getName().equals(next.getName()))
			return false;
		
		Class[] fargs = next.getParameterTypes();
		if(fargs.length!=this.getVariableCount())
			return false;
		
		
		Iterator vars = this.getVariableTypes();
		for(int i=0;i<fargs.length;++i) {
			String type = vars.next().toString();
			String expectedType = fargs[i].getName();
			if(!type.equals(expectedType))
				return false;
		}
		
		String expectedReturnType = next.getReturnType().getName();
		if(!expectedReturnType.equals(this.getReturnType()))
			return false;
		
		return true;
	}

}


